package de.intevation.flys.exports;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.Range;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.TextAnchor;

import org.w3c.dom.Document;

import de.intevation.artifacts.Artifact;

import de.intevation.artifactdatabase.state.Facet;

import de.intevation.flys.artifacts.FLYSArtifact;

import de.intevation.flys.artifacts.model.FacetTypes;
import de.intevation.flys.artifacts.model.WQKms;

import de.intevation.flys.model.Annotation;
import de.intevation.flys.jfree.StickyAxisAnnotation;


/**
 * An OutGenerator that generates discharge curves.
 *
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class LongitudinalSectionGenerator
extends      XYChartGenerator
implements   FacetTypes
{
    /** The logger that is used in this generator. */
    private static Logger logger =
        Logger.getLogger(LongitudinalSectionGenerator.class);

    public static final String I18N_CHART_TITLE =
        "chart.longitudinal.section.title";

    public static final String I18N_ANNOTATIONS_LABEL =
        "chart.longitudinal.annotations.label";

    public static final String I18N_CHART_SUBTITLE =
        "chart.longitudinal.section.subtitle";

    public static final String I18N_XAXIS_LABEL =
        "chart.longitudinal.section.xaxis.label";

    public static final String I18N_YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.label";

    public static final String I18N_2YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.second.label";

    public static final String I18N_CHART_TITLE_DEFAULT  = "W-L\u00e4ngsschnitt";
    public static final String I18N_XAXIS_LABEL_DEFAULT  = "km";
    public static final String I18N_YAXIS_LABEL_DEFAULT  = "W [NN + m]";
    public static final String I18N_2YAXIS_LABEL_DEFAULT = "Q [m\u00b3/s]";


    protected boolean inverted;

    /** List of annotations to insert in plot. */
    protected List<Annotation> annotations;

    /** Pseudo-Dataseries to have a legend for the annotations. */
    protected XYSeriesCollection pseudoAnnotationData = null;


    public LongitudinalSectionGenerator() {
        super();
        annotations = new ArrayList<Annotation>();
    }


    protected String getChartTitle() {
        return msg(I18N_CHART_TITLE, I18N_CHART_TITLE_DEFAULT);
    }

    public boolean isInverted() {
        return inverted;
    }

    public void setInverted(boolean inverted) {
        this.inverted = inverted;
    }

    @Override
    protected void addSubtitles(JFreeChart chart) {
        double[] dist  = getRange();

        Object[] args = new Object[] {
            getRiverName(),
            dist[0],
            dist[1]
        };

        String subtitle = msg(I18N_CHART_SUBTITLE, "", args);
        chart.addSubtitle(new TextTitle(subtitle));
    }

    @Override
    public JFreeChart generateChart() {
        JFreeChart c = super.generateChart();
        XYPlot p = (XYPlot) c.getPlot();

        redoAnnotations(p, p.getDomainAxis());

        return c;
    }


    protected String getXAxisLabel() {
        return msg(I18N_XAXIS_LABEL, I18N_XAXIS_LABEL_DEFAULT);
    }


    protected String getYAxisLabel() {
        return msg(I18N_YAXIS_LABEL, I18N_YAXIS_LABEL_DEFAULT);
    }


    protected void adjustAxes(XYPlot plot) {
        super.adjustAxes(plot);

        NumberAxis qAxis = new NumberAxis(
            msg(I18N_2YAXIS_LABEL, I18N_2YAXIS_LABEL_DEFAULT));

        plot.setRangeAxis(1, qAxis);

        invertXAxis(plot.getDomainAxis());
    }


    /**
     * Remove all annotations from plot and re-insert them at an approximately
     * okay position. The followed approach is naive but side-effect free.
     *
     * @param plot the plot.
     * @param axis the value axis.
     */
    protected void redoAnnotations(XYPlot plot, ValueAxis axis) {
        plot.clearAnnotations();
        // TODO Position calculation could/should be done in
        // the StickyAxisAnnotation-Implementation itself.
        ValueAxis yAxis = plot.getRangeAxis();
        float posY = 140.f;
        if (yAxis != null) {
            posY = (float) yAxis.getRange().getLowerBound();
            // Add some (2%) space between Text and axis.
            posY += 0.02f * (yAxis.getRange().getUpperBound()
                    - yAxis.getRange().getLowerBound());
        }

        // Add all annotations.
        for (Annotation a: annotations) {
            float posX = (float) a.getRange().getA().doubleValue();
            String text = a.getPosition().getValue();

            StickyAxisAnnotation ta = new StickyAxisAnnotation(text, posX, posY);
            double rotation = 270.0f * (Math.PI / 180.0f);
            ta.setRotationAngle(rotation);
            ta.setRotationAnchor(TextAnchor.CENTER_LEFT);
            ta.setTextAnchor(TextAnchor.CENTER_LEFT);
            plot.getRenderer().addAnnotation(ta);
        }
    }


    /**
     * Create a range that includes 0 (for the Q axis).
     * @param range range with which to look up upper bound.
     * @return range with 0 included.
     */
    protected Range createSecondAxisRange(Range range) {
       return new Range(0d, range.getUpperBound());
    }


    /**
     * This method overrides the XYChartGenerators zoomY method to be able to
     * modify the range of the Q axis (here, it shall include 0).
     */
    @Override
    protected boolean zoomY(XYPlot plot, ValueAxis axis, Range range, Range x) {
        if (plot.getRangeAxisIndex(axis) == 1) {
            // We want the Q axis to start at 0 if no zooming has been done.
            range = createSecondAxisRange(range);
        }

        return super.zoomY(plot, axis, range, x);
    }


    /**
     * This method inverts the x-axis based on the kilometer information of the
     * selected river. If the head of the river is at kilometer 0, the axis is
     * not inverted, otherwise it is.
     *
     * @param xaxis The domain axis.
     */
    protected void invertXAxis(ValueAxis xaxis) {

        if (inverted) {
            logger.debug("Invert X-Axis.");
            xaxis.setInverted(true);
        }
    }


    public void doOut(Artifact artifact, Facet facet, Document attr) {
        String name = facet.getName();

        logger.debug("LongitudinalSectionGenerator.doOut: " + name);

        if (name == null) {
            logger.error("No facet name for doOut(). No output generated!");
            return;
        }

        FLYSArtifact flys = (FLYSArtifact) artifact;
        Facet        f    = flys.getNativeFacet(facet);

        if (f == null) {
            return;
        }

        if (name.equals(LONGITUDINAL_W)) {
            doWOut((WQKms) f.getData(artifact, context), attr);
        }
        else if (name.equals(LONGITUDINAL_Q)) {
            doQOut((WQKms) f.getData(artifact, context), attr);
        }
        else if (name.equals(LONGITUDINAL_ANNOTATION)) {
            doAnnotationsOut(f.getData(artifact, context), attr);
        }
        else {
            logger.warn("Unknown facet name: " + name);
            return;
        }
    }


    /**
     * Register annotations available for the diagram.
     *
     * @param o     list of annotations (data of facet).
     * @param theme yet ignored.
     */
    protected void doAnnotationsOut(Object o, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doAnnotationsOut");
        this.annotations = (List<Annotation>) o;

        // Add (empty) pseudo series to have legend entry for annotations.
        if (pseudoAnnotationData == null) {
            pseudoAnnotationData = new XYSeriesCollection();
        }
        // Localized legend.
        Object[] args = new Object[] {getRiverName()};
        String label = msg(I18N_ANNOTATIONS_LABEL, "", args);
        pseudoAnnotationData.addSeries(new StyledXYSeries(label, theme));
    }


    /**
     * Process the output for W facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     * @param theme The theme that contains styling information.
     */
    protected void doWOut(WQKms wqkms, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doWOut");

        XYSeries series = new StyledXYSeries(getSeriesName(wqkms, "W"), theme);

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKm(0));
                logger.debug("End   km: " + wqkms.getKm(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKm(i), wqkms.getW(i));
        }

        addFirstAxisSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    /**
     * Process the output for Q facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     * @param theme The theme that contains styling information.
     */
    protected void doQOut(WQKms wqkms, Document theme) {
        logger.debug("LongitudinalSectionGenerator.doQOut");

        XYSeries series = new StyledXYSeries(getSeriesName(wqkms, "Q"), theme);

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKm(0));
                logger.debug("End   km: " + wqkms.getKm(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKm(i), wqkms.getQ(i));
        }

        addSecondAxisSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    /**
     * Add datasets to plot.
     * @param plot plot to add datasets to.
     */
    @Override
    protected void addDatasets(XYPlot plot) {
        super.addDatasets(plot);
        if (pseudoAnnotationData != null) {
            plot.setDataset(2, pseudoAnnotationData);
        }
    }


    /**
     * Get name of series (displayed in legend).
     * @return name of the series.
     */
    protected String getSeriesName(WQKms wqkms, String mode) {
        String name   = wqkms.getName();
        String prefix = name != null && name.indexOf(mode) >= 0 ? null : mode;

        return prefix != null && prefix.length() > 0
            ? prefix + "(" + name +")"
            : name;
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
